@@ -219,6 +219,31 @@ class TestFormatting:
         assert "\n" not in newlines
         assert "\t" not in tabs
 
+    def test_index_repr(self):
+        from xarray.core.indexes import Index
+
+        class CustomIndex(Index):
+            def __init__(self, names):
+                self.names = names
+
+            def __repr__(self):
+                return f"CustomIndex(coords={self.names})"
+
+        coord_names = ["x", "y"]
+        index = CustomIndex(coord_names)
+        name = "x"
+
+        normal = formatting.summarize_index(name, index, col_width=20)
+        assert name in normal
+        assert "CustomIndex" in normal
+
+        CustomIndex._repr_inline_ = (
+            lambda self, max_width: f"CustomIndex[{', '.join(self.names)}]"
+        )
+        inline = formatting.summarize_index(name, index, col_width=20)
+        assert name in inline
+        assert index._repr_inline_(max_width=40) in inline
+
     def test_diff_array_repr(self) -> None:
         da_a = xr.DataArray(
             np.array([[1, 2, 3], [4, 5, 6]], dtype="int64"),
@@ -550,17 +575,28 @@ def test_large_array_repr_length() -> None:
 
 @requires_netCDF4
 def test_repr_file_collapsed(tmp_path) -> None:
-    arr = xr.DataArray(np.arange(300), dims="test")
-    arr.to_netcdf(tmp_path / "test.nc", engine="netcdf4")
+    arr_to_store = xr.DataArray(np.arange(300, dtype=np.int64), dims="test")
+    arr_to_store.to_netcdf(tmp_path / "test.nc", engine="netcdf4")
 
     with xr.open_dataarray(tmp_path / "test.nc") as arr, xr.set_options(
         display_expand_data=False
     ):
-        actual = formatting.array_repr(arr)
+        actual = repr(arr)
         expected = dedent(
             """\
         <xarray.DataArray (test: 300)>
-        array([  0,   1,   2, ..., 297, 298, 299])
+        [300 values with dtype=int64]
+        Dimensions without coordinates: test"""
+        )
+
+        assert actual == expected
+
+        arr_loaded = arr.compute()
+        actual = arr_loaded.__repr__()
+        expected = dedent(
+            """\
+        <xarray.DataArray (test: 300)>
+        0 1 2 3 4 5 6 7 8 9 10 11 12 ... 288 289 290 291 292 293 294 295 296 297 298 299
         Dimensions without coordinates: test"""
         )
 
@@ -637,12 +673,12 @@ def test__mapping_repr_recursive() -> None:
     # GH:issue:7111
 
     # direct recursion
-    ds = xr.Dataset({"a": [["x"], [1, 2, 3]]})
+    ds = xr.Dataset({"a": ("x", [1, 2, 3])})
     ds.attrs["ds"] = ds
     formatting.dataset_repr(ds)
 
     # indirect recursion
-    ds2 = xr.Dataset({"b": [["y"], [1, 2, 3]]})
+    ds2 = xr.Dataset({"b": ("y", [1, 2, 3])})
     ds.attrs["ds"] = ds2
     ds2.attrs["ds"] = ds
     formatting.dataset_repr(ds2)
@@ -674,3 +710,18 @@ def test__element_formatter(n_elements: int = 100) -> None:
     )
     actual = intro + values
     assert expected == actual
+
+
+def test_lazy_array_wont_compute() -> None:
+    from xarray.core.indexing import LazilyIndexedArray
+
+    class LazilyIndexedArrayNotComputable(LazilyIndexedArray):
+        def __array__(self, dtype=None):
+            raise NotImplementedError("Computing this array is not possible.")
+
+    arr = LazilyIndexedArrayNotComputable(np.array([1, 2]))
+    var = xr.DataArray(arr)
+
+    # These will crash if var.data are converted to numpy arrays:
+    var.__repr__()
+    var._repr_html_()
